body {
  margin: 0em;
  padding: 0px;
  color: rgb(0,0,0);
  background: #ffffff;
  font-family: LibvirtOverpass;
}

p, ul, ol, dl {
  padding: 0px;
  margin: 0px;
}

p {
  margin-top: 1em;
  margin-bottom: 1em;
}

ul, ol {
  margin-left: 2em;
}

dt {
  margin-left: 1em;
  margin-right: 2em;
}

dt code, dt .literal {
  font-weight: bold;
}

dl dd {
  margin-left: 2em;
  margin-right: 2em;
  margin-bottom: 0.5em;
}

h1, h2, h3, h4, h5, h6 {
  font-weight: bold;
  margin: 0px;
  padding: 0px;
  margin-bottom: 0.25em;
}

h1 {
  margin-top: 0em;
  font-size: 1.6em;
}

h2 {
  margin-top: 1.0em;
  font-size: 1.4em;
}

h3 {
  margin-top: 1.0em;
  font-size: 1.2em;
}

h4 {
  margin-top: 1.0em;
  font-size: 1.1em;
}

h5 {
  margin-top: 0.75em;
  font-size: 1em;
}

h6 {
  margin-top: 0.75em;
  font-size: 0.8em;
}

code, pre, tt {
  font-family: LibvirtOverpassMono;
}

dd code, p code, tt {
  background-color: #eeeeee;
}

pre {
  font-size: 90%;
}
